from __future__ import annotations
import threading


class SingletonVars(object):
    _instance: object | None
    _lock: threading.Lock
    _is_initialized: bool
    _type: type

    def __init__(self, t: type) -> None:
        self._instance = None
        self._lock = threading.Lock()
        self._is_initialized = False
        self._type = t

    @property
    def instance(self) -> object | None:
        return self._instance

    @instance.setter
    def instance(self, obj: object) -> None:
        if self._instance is None:
            self._instance = obj

    @property
    def lock(self) -> threading.Lock:
        return self._lock

    @property
    def is_initialized(self) -> bool:
        return self._is_initialized

    @is_initialized.setter
    def is_initialized(self, value: bool) -> None:
        if value:
            self._is_initialized = True

    @property
    def t(self) -> type:
        return self._type



from __future__ import annotations
from typing import ClassVar
from .singleton_vars import SingletonVars


class _singletonreg_mcls(type):
    _data: ClassVar[dict[type, SingletonVars]]

    @property
    def data(cls) -> dict[type, SingletonVars]:
        mcls = type(cls)
        if not hasattr(mcls, "_data"):
            setattr(mcls, "_data", {})
        return getattr(mcls, "_data")

    def is_singleton(cls, t: type) -> bool:
        return cls.data.get(t) is not None


class SingletonRegistry(object, metaclass=_singletonreg_mcls):
    pass



from __future__ import annotations
import heresy


def test_singletonproperty() -> None:
    @heresy.singleton
    class Sample(object):
        _value: int
        def __init__(self):
            self._value = 0
        @heresy.singletonproperty
        def value(self) -> int: # pyright: ignore[reportRedeclaration]
            return self._value
        @value.setter
        def value(self, v: int) -> None:
            self._value = v
    Sample.value = 5
    assert(Sample.value == 5)



from __future__ import annotations
import heresy


def test_singleton() -> None:
    @heresy.singleton
    class Sample(object):
        def __init__(self):
            pass
    a = Sample()
    b = Sample()
    assert(a is b)



from __future__ import annotations
from typing import TypeVar, Any, cast
import inspect
from ._internal.singleton_vars import SingletonVars
from ._internal.singleton_registry import SingletonRegistry


_T = TypeVar("_T")


def singleton(_cls: type[_T]) -> type[_T]:
    if len(inspect.signature(_cls.__init__).parameters) != 1:
        raise TypeError(f"{_cls.__name__}.__init__ must only accept 1 argument: self.")

    if _cls not in SingletonRegistry.data:
        SingletonRegistry.data[_cls] = SingletonVars(_cls)
    else:
        raise TypeError("This class is already a singleton.")

    orig_new = _cls.__new__
    orig_init = _cls.__init__

    def __new__(cls: type[_T], *args: Any, **kwargs: Any) -> _T:
        reg_data = SingletonRegistry.data[cls]
        if reg_data.instance is None:
            with reg_data.lock:
                if reg_data.instance is None:
                    reg_data.instance = orig_new(cls, *args, **kwargs)
        return cast(_T, reg_data.instance)

    def __init__(self: object) -> None:
        reg_data = SingletonRegistry.data[type(self)]
        if not reg_data.is_initialized:
            orig_init(cast(_T, self))
            reg_data.is_initialized = True


    setattr(_cls, "__new__", staticmethod(__new__))
    setattr(_cls, "__init__", __init__)

    return _cls



from __future__ import annotations
from typing import TypeVar, Any, Generic
import collections.abc as cabc
import threading
from ._internal.singleton_registry import SingletonRegistry


_T = TypeVar("_T")
_R = TypeVar("_R")


class singletonproperty(Generic[_T, _R]):
    _fget: cabc.Callable[[_T], _R] | None
    _fset: cabc.Callable[[_T, Any], None] | None
    _fdel: cabc.Callable[[_T], None] | None
    _owner: type[_T]
    _lock: threading.Lock

    def __set_name__(self, owner: type[_T], _: str) -> None:
        self._owner = owner

    def __init__(
        self,
        fget: cabc.Callable[[_T], _R] | None = None,
        fset: cabc.Callable[[_T, Any], None] | None = None,
        fdel: cabc.Callable[[_T], None] | None = None,
        doc: str | None = None
    ) -> None:
        self._fget = fget
        self._fset = fset
        self._fdel = fdel
        self._lock = threading.Lock()
        self.__doc__ = doc if doc is not None else (fget.__doc__ if fget else None)

    @property
    def _instance(self) -> _T:
        if self._owner not in SingletonRegistry.data.keys():
            raise TypeError("singletonproperty must only be used inside a singleton.")
        return self._owner()

    def __get__(self, obj: _T | None, _: type[_T] | None = None) -> _R:
        if self._fget is None:
            raise AttributeError("cannot get attribute.")
        with self._lock:
            return self._fget(self._instance)

    def __set__(self, _: object, value: Any) -> None:
        if self._fset is None:
            raise AttributeError("cannot set attribute.")
        with self._lock:
            return self._fset(self._instance, value)

    def __delete__(self, _: object) -> None:
        if self._fdel is None:
            raise AttributeError("cannot delete attribute.")
        with self._lock:
            self._fdel(self._instance)

    def getter(self, fget: cabc.Callable[[_T], _R]) -> singletonproperty[_T, _R]:
        self._fget = fget
        return self

    def setter(self, fset: cabc.Callable[[_T, Any], None]) -> singletonproperty[_T, _R]:
        self._fset = fset
        return self

    def deleter(self, fdel: cabc.Callable[[_T], None]) -> singletonproperty[_T, _R]:
        self._fdel = fdel
        return self



from __future__ import annotations
from typing import TypeVar, ParamSpec, Generic, Concatenate
import collections.abc as cabc
import functools
from ._internal.singleton_registry import SingletonRegistry


_T = TypeVar("_T")
_R = TypeVar("_R")
_P = ParamSpec("_P")


class singletonmethod(Generic[_T, _P, _R]):
    _func: cabc.Callable[Concatenate[_T, _P], _R]

    def __init__(self, func: cabc.Callable[Concatenate[_T, _P], _R]) -> None:
        self._func = func

    def __get__(self, _: _T | None, owner: type[_T]) -> cabc.Callable[_P, _R]:
        if owner not in SingletonRegistry.data.keys():
            raise TypeError("singletonproperty must only be used inside a singleton.")
        @functools.wraps(self._func)
        def method(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            instance = owner()
            return self._func(instance, *args, **kwargs)
        return method



from __future__ import annotations
from ._version import __version__
from ._root_submodule.singleton.singleton import singleton
from ._root_submodule.singleton.singletonproperty import singletonproperty
from ._root_submodule.singleton.singletonmethod import singletonmethod

__all__: list[str] = ["__version__", "singleton", "singletonproperty", "singletonmethod"]
